package
{
	import flash.utils.getTimer;

	public class AStar
	{
//		private var _open:Array = [];
		private var _open:Binary;
//		private var _closed:Array;
		private var _closed:Array;
		private var _grid:Grid;
		private var _endNode:Node;
		private var _startNode:Node;
		private var _path:Array;
//		private var _heuristic:Function = manhattan;
//		private var _heuristic:Function = euclidian;
		private var _heuristic:Function = diagonal;
		private var _straightCost:Number = 1.0;
		private var _diagCost:Number = Math.SQRT2;
		
		public function AStar()
		{
		}
		
		public function findPath(grid:Grid):Boolean
		{
			_grid = grid;
//			_open = new Array();
			_open = new Binary("f");
			_closed = new Array();
			
			_startNode = _grid.startNode;
			_endNode = _grid.endNode;
			
			_startNode.g = 0;
			_startNode.h = _heuristic(_startNode);
			_startNode.f = _startNode.g + _startNode.h;
			
			return search();
		}
		
		public function search():Boolean
		{
			var startTime:int=getTimer();
			var node:Node = _startNode;
			var sortTime:int = 0;
			var tryCount:int = 0;
			while(node != _endNode)
			{
				tryCount++;
				var startX:int = 0 > node.x - 1 ? 0 : node.x - 1;
				var endX:int = _grid.numCols - 1 < node.x + 1 ? _grid.numCols - 1 : node.x + 1;
				var startY:int = 0 > node.y - 1 ? 0 : node.y - 1;
				var endY:int = _grid.numRows - 1 < node.y + 1 ? _grid.numRows - 1 : node.y + 1;
				
				
				for(var i:int = startX; i <= endX; i++)
				{
					for(var j:int = startY; j <= endY; j++)
					{
						var test:Node = _grid.getNode(i, j);
						if(test == node)
						{
							continue;
						}
						
						if( !test.walkable ||
							!_grid.getNode(node.x, test.y).walkable ||
							!_grid.getNode(test.x, node.y).walkable )
						{
							test.costMultiplier = 999;
						}
						else
						{
							test.costMultiplier = 1;
						}
							
						var cost:Number = _straightCost;
						
						if(!((node.x == test.x) || (node.y == test.y)))
						{
							cost = _diagCost;
						}
						
						var g:Number = node.g + cost * test.costMultiplier;
						var h:Number = _heuristic(test);
						var f:Number = g + h;
						var isInOpen:Boolean = _open.indexOf(test) != -1;
						if( isInOpen || _closed.indexOf(test) != -1)
						{
							if(test.f > f)
							{
								test.f = f;
								test.g = g;
								test.h = h;
								test.parent = node;
								if( isInOpen )
									_open.updateNode( test );
							}
						}
						else
						{
							test.f = f;
							test.g = g;
							test.h = h;
							test.parent = node;
//							_open.push(test);
							_open.push( test );
						}
					}
				}
				_closed.push(node);
				if(_open.length == 0)
				{
					trace("no path found");
					
					return false
				}
//				var sortStartTime:int = getTimer();
//				_open.sortOn("f", Array.NUMERIC);
//				sortTime += (getTimer() - sortStartTime);
//				node = _open.shift() as Node;
				node = _open.shift() as Node;
			}
			trace( "time cost: " + (getTimer() - startTime) + "ms");
//			trace( "sort cost: " + sortTime);
			trace( "try time: " + tryCount);
			buildPath();
			return true;
		}
		
		private function buildPath():void
		{
			_path = new Array();
			var node:Node = _endNode;
			_path.push(node);
			
			//不包含起始节点
			while(node.parent != _startNode)
			{
				node = node.parent;
				_path.unshift(node);
			}
			//排除无法移动点
			var len:int = _path.length;
			for( var i:int=len-1; i>=0; i-- )
			{
				if( _path[i].walkable == false )
				{
					_path.splice(i, len-i);
					break;
				}
				else if( len == 1 && !isDiagonalWalkable(_startNode, _endNode) )
				{
					_path.shift();
				}
				else if( i < len - 1 && !isDiagonalWalkable(_path[i], _path[i+1]) )
				{
					_path.splice(i+1, len-i-1);
					break;
				}
			}
		}
		
		public function get path():Array
		{
			return _path;
		}
//		
//		/** 向开启列表中添加元素 */
//		public function addToOpenList( node:Node ):void
//		{
//			//将新节点添至末尾先
//			_open.push( node );
//			var len:int = _open.length;
//			
//			//若数组中只有一个元素则省略排序过程，否则对新元素执行上浮过程
//			if( len > 1 )
//			{
//				/** 新添入节点当前所在索引 */
//				var index:int = len;
//				
//				/** 新节点当前父节点所在索引 */
//				var parentIndex:int = index / 2 - 1;
//				
//				var temp:Node;
//				
//				//和它的父节点（位置为当前位置除以2取整，比如第4个元素的父节点位置是2，第7个元素的父节点位置是3）比较，
//				//如果新元素比父节点元素小则交换这两个元素，然后再和新位置的父节点比较，直到它的父节点不再比它大，
//				//或者已经到达顶端，及第1的位置
//				while( _open[parentIndex].f > node.f )
//				{
//					temp = _open[parentIndex];
//					_open[parentIndex] = node;
//					_open[index - 1] = temp;
//					index /= 2;
//					parentIndex = index / 2 - 1;
//				}
//			}
//			
//		}
//		
//		/** 弹出开启列表中第一个元素 */
//		public function shiftFromOpenList():Node
//		{
//			//先弹出列首元素
//			var result:Node =  _open.shift();
//			
//			/** 数组长度 */
//			var len:int = _open.length;
//			
//			//若弹出列首元素后数组空了或者其中只有一个元素了则省略排序过程，否则对列尾元素执行下沉过程
//			if( len > 1 )
//			{
//				/** 列尾节点 */
//				var lastNode:Node = _open.pop();
//				
//				//将列尾元素排至首位
//				_open.unshift( lastNode );
//				
//				/** 末尾节点当前所在索引 */
//				var index:int = 0;
//				
//				/** 末尾节点当前第一子节点所在索引 */
//				var childIndex:int = (index + 1) * 2 - 1;
//				
//				/** 末尾节点当前两个子节点中较小的一个的索引 */
//				var comparedIndex:int;
//				
//				var temp:Node;
//				
//				//和它的两个子节点比较，如果较小的子节点比它小就将它们交换，直到两个子节点都比它大
//				while( childIndex < len )
//				{
//					//只有一个子节点的情况
//					if( childIndex + 1 == len )
//					{
//						comparedIndex = childIndex;
//					}
//					//有两个子节点则取其中较小的那个
//					else
//					{
//						comparedIndex = _open[childIndex].f < _open[childIndex + 1].f ? childIndex : childIndex + 1;
//					}
//					
//					if( _open[comparedIndex].f < lastNode.f )
//					{
//						temp = _open[comparedIndex];
//						_open[comparedIndex] = lastNode;
//						_open[index] = temp;
//						index = comparedIndex;
//						childIndex = (index + 1) * 2 - 1;
//					}
//					else
//					{
//						break;
//					}
//				}
//				
//			}
//
//			
//			return result;
//		}

		
		/** 判断两个节点的对角线路线是否可走 */
		private function isDiagonalWalkable( node1:Node, node2:Node ):Boolean
		{
			var nearByNode1:Node = _grid.getNode( node1.x, node2.y );
			var nearByNode2:Node = _grid.getNode( node2.x, node1.y );
			
			if( nearByNode1.walkable && nearByNode2.walkable )return true;
			return false;
		}
		
		private function manhattan(node:Node):Number
		{
			return Math.abs(node.x - _endNode.x) * _straightCost + Math.abs(node.y + _endNode.y) * _straightCost;
		}
		
		private function euclidian(node:Node):Number
		{
			var dx:Number = node.x - _endNode.x;
			var dy:Number = node.y - _endNode.y;
			return Math.sqrt(dx * dx + dy * dy) * _straightCost;
		}
		
		private function diagonal(node:Node):Number
		{
			var dx:Number = node.x - _endNode.x < 0 ? _endNode.x - node.x : node.x - _endNode.x;
			var dy:Number = node.y - _endNode.y < 0 ? _endNode.y - node.y : node.y - _endNode.y;
			var diag:Number = dx < dy ? dx : dy;
			var straight:Number = dx + dy;
			return _diagCost * diag + _straightCost * (straight - 2 * diag);
		}
		
//		public function get visited():Vector.<Node>
//		{
//			return _closed.concat(_open);
//		}
	}
}
